/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*-
 *
 * Copyright (C) 2017 Mohammed Sadiq <sadiq@sadiqpk.org>
 * Copyright (C) 2010 Red Hat, Inc
 * Copyright (C) 2008 William Jon McCann <jmccann@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <config.h>

#include <glib.h>
#include <glib/gi18n.h>

#include "list-box-helper.h"
#include "cc-font-panel.h"
#include "cc-font-resources.h"

/* font options */
#define FONT_NAME                "font-name"
#define MONOSPACE_FONT_NAME      "monospace-font-name"
#define CDOS_INTERFACE_SCHEMA    "org.cdos.desktop.interface"

struct _CcFontPanel
{
  CcPanel              parent_instance;
  GtkWidget     *font_list_box;
  GtkFontButton *default_font;
  GtkFontButton *monospace_font;

  GSettings *settings;
};

CC_PANEL_REGISTER (CcFontPanel, cc_font_panel)

static void
default_font_changed (GtkFontButton *button, CcFontPanel * self) 
{

  gchar * select_font = gtk_font_chooser_get_font (button);
  if (self->settings != NULL) 
  {
    g_settings_set_string (self->settings,
                         FONT_NAME, select_font);
  }

  g_free (select_font);
}

static void
monospace_font_changed (GtkFontButton *button, CcFontPanel * self) 
{
  gchar * select_font = gtk_font_chooser_get_font (button);
  if (self->settings != NULL) 
  {
    g_settings_set_string (self->settings, MONOSPACE_FONT_NAME, select_font);
  }
  g_free(select_font);
}

static void
font_button_set_default_font_name (CcFontPanel *self, GtkWidget *button, const gchar *key)
{
  gchar *font_name;
  font_name = g_settings_get_string (self->settings, key);
  gtk_font_button_set_font_name (button, font_name);
  g_free (font_name);
}

static void
font_button_set_monospace_font_name (CcFontPanel *self, GtkWidget *button, const gchar *key)
{
  gchar *font_name;
  font_name = g_settings_get_string (self->settings, key);
  gtk_font_button_set_font_name (button, font_name);
  g_free (font_name);
}

static void
info_panel_setup_media (CcFontPanel *self)
{
  font_button_set_default_font_name (self, self->default_font, FONT_NAME);
  font_button_set_monospace_font_name (self, self->monospace_font, MONOSPACE_FONT_NAME);

  g_signal_connect (self->default_font, 
                    "font_set", 
                    G_CALLBACK(default_font_changed), 
                    self);
  g_signal_connect (self->monospace_font, 
                    "font_set", 
                    G_CALLBACK(monospace_font_changed), 
                    self);
}

static void
cc_font_panel_finalize (GObject *object)
{
  CcFontPanel *self = CC_FONT_PANEL (object);

  G_OBJECT_CLASS (cc_font_panel_parent_class)->finalize (object);
}

static void
cc_font_panel_dispose (GObject *object)
{
  CcFontPanel *self = CC_FONT_PANEL (object);

  g_clear_object (&self->settings);

  G_OBJECT_CLASS (cc_font_panel_parent_class)->dispose (object);
}

static void
font_reset_default_settings (CcFontPanel *self)
{
  g_settings_reset (self->settings, FONT_NAME);
  g_settings_reset (self->settings, MONOSPACE_FONT_NAME);
}

static void
cdos_update_active (GSettings   *settings,
                    const gchar  *key,
                    CcFontPanel  *self)
{
  gchar *font = NULL;
  if (g_str_equal (key, FONT_NAME))
  {
    font = g_settings_get_string (settings, key);
    gtk_font_chooser_set_font (self->default_font, font);
    g_free (font);
  }
  else if (g_str_equal (key, MONOSPACE_FONT_NAME))
  {
    font = g_settings_get_string (settings, key);
    gtk_font_chooser_set_font (self->monospace_font, font);
    g_free (font);
  }
}

static void
cc_font_panel_class_init (CcFontPanelClass *klass)
{
  GObjectClass   *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  object_class->finalize = cc_font_panel_finalize;
  object_class->dispose = cc_font_panel_dispose;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/control-center/font/cc-font-panel.ui");

  gtk_widget_class_bind_template_child (widget_class, CcFontPanel, font_list_box);
  gtk_widget_class_bind_template_child (widget_class, CcFontPanel, default_font);
  gtk_widget_class_bind_template_child (widget_class, CcFontPanel, monospace_font);

  gtk_widget_class_bind_template_callback (widget_class, font_reset_default_settings);
}

static void
cc_font_panel_init (CcFontPanel *self)
{
  g_resources_register (cc_font_get_resource ());
  gtk_widget_init_template (GTK_WIDGET (self));

  gtk_list_box_set_header_func (self->font_list_box,
                                cc_list_box_update_header_func,
                                NULL, NULL);
  self->settings  = g_settings_new (CDOS_INTERFACE_SCHEMA);
  info_panel_setup_media (self);
  g_signal_connect(self->settings, 
                   "changed",
                   G_CALLBACK(cdos_update_active), 
                   self);
}
