/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*-
 *
 * Copyright (C) 2019 Purism SPC
 * Copyright (C) 2017 Mohammed Sadiq <sadiq@sadiqpk.org>
 * Copyright (C) 2010 Red Hat, Inc
 * Copyright (C) 2008 William Jon McCann <jmccann@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <config.h>

#include <glib.h>
#include <glib/gi18n.h>
#include "cc-util.h"

#include "cc-performance-mode-panel.h"
#include "cc-performance-mode-resources.h"
#define SUPPORT_KEY                "support-performance-mode"
#define DESKTOP_SESSION_SCHEMA          "org.cdos.desktop.session"
#define PERFORMANCE_MODE          "mate"
#define ORIGINAL_MODE             "lightdm-xsession"

struct _CcPerformanceModePanel
{
  CcPanel     parent_instance;

  GtkLabel   *attention_lab;
  GtkSwitch   *switch_btn;

  gboolean     button_state;
  GDBusProxy  *account_proxy;
  GDBusProxy  *account_user_proxy;
  gchar     *user_path;

};

G_DEFINE_TYPE (CcPerformanceModePanel, cc_performance_mode_panel, CC_TYPE_PANEL)


static void
performance_call_dbus_method ( CcPerformanceModePanel  *self, gchar *session_type)
{
  g_autoptr(GDBusProxy)  user_proxy = NULL;
  g_autoptr(GError) error = NULL;
  g_autoptr(GVariant) res = NULL;
  user_proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SYSTEM,
                                                G_DBUS_PROXY_FLAGS_NONE,
                                                NULL,
                                                "org.freedesktop.Accounts",
                                                self->user_path,
                                                "org.freedesktop.Accounts.User",
                                                NULL,
                                                &error);
  if (error != NULL)
  {
      g_warning ("%s", error->message);
      return;
  }
  res = g_dbus_proxy_call_sync (user_proxy,
                                          "SetXSession",
                                          g_variant_new ("(s)",
                                                        session_type),
                                          G_DBUS_CALL_FLAGS_NONE,
                                          -1,
                                          NULL,
                                          &error);
  if (error != NULL)
  {
      g_warning ("%s", error->message);
      return;
  }
}

static void
switch_changed_cb (CcPerformanceModePanel  *self)
{
  gboolean state;
  state = gtk_switch_get_active (self->switch_btn);
  if (state)
    {
      performance_call_dbus_method(self, PERFORMANCE_MODE);
    }
  else
    {
      performance_call_dbus_method(self, ORIGINAL_MODE);
    }
    self->button_state = state;
    gtk_label_set_text (self->attention_lab, _("Attention: change setting need  logout or restart"));
}

static void
cc_performance_mode_panel_finalize (GObject *object)
{
  CcPerformanceModePanel  *self = CC_PERFORMANCE_MODE_PANEL (object);
  g_free (self->user_path);
  G_OBJECT_CLASS (cc_performance_mode_panel_parent_class)->finalize (object);
}

static void
cc_performance_mode_panel_dispose (GObject *object)
{
  CcPerformanceModePanel  *self = CC_PERFORMANCE_MODE_PANEL (object);
  g_clear_object (&self->account_proxy);
  g_clear_object (&self->account_user_proxy);
  G_OBJECT_CLASS (cc_performance_mode_panel_parent_class)->dispose (object);
}

static void
cc_performance_mode_panel_class_init (CcPerformanceModePanelClass *klass)
{
  GObjectClass    *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass  *widget_class = GTK_WIDGET_CLASS (klass);

  object_class->finalize = cc_performance_mode_panel_finalize;
  object_class->dispose = cc_performance_mode_panel_dispose;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/control-center/performance-mode/cc-performance-mode-panel.ui");

  gtk_widget_class_bind_template_child (widget_class, CcPerformanceModePanel, attention_lab);
  gtk_widget_class_bind_template_child (widget_class, CcPerformanceModePanel, switch_btn);
  
  gtk_widget_class_bind_template_callback (widget_class, switch_changed_cb);

}

static void
load_custom_css (CcPerformanceModePanel *self)
{
  g_autoptr(GtkCssProvider) provider = NULL;

  provider = gtk_css_provider_new ();
  gtk_css_provider_load_from_resource (provider, "/org/gnome/control-center/performance-mode/perf.css");
  gtk_style_context_add_provider_for_screen (gdk_screen_get_default (),
                                             GTK_STYLE_PROVIDER (provider),
                                             GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
}

static void
cc_performance_mode_panel_init (CcPerformanceModePanel  *self)
{
  g_resources_register (cc_performance_mode_get_resource ());
  load_custom_css (self);
  gtk_widget_init_template (GTK_WIDGET (self));

  g_autoptr(GError) error = NULL;
  g_autoptr(GVariant) res = NULL;
  g_autoptr(GVariant) res_session = NULL;

  g_autoptr(GVariant) inner = NULL;
  g_autoptr(GCancellable) cancellable = NULL;
  g_autoptr(GSettings) session_support_settings = NULL;

  const gchar *opath = NULL;
  const gchar *user_name = NULL;
  const gchar *user_session = NULL;
  gboolean support_enabled = FALSE;

  user_name = g_get_user_name();
  cancellable = g_cancellable_new ();
  self->account_proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SYSTEM,
                                                G_DBUS_PROXY_FLAGS_NONE,
                                                NULL,
                                                "org.freedesktop.Accounts",
                                                "/org/freedesktop/Accounts",
                                                "org.freedesktop.Accounts",
                                                cancellable,
                                                &error);
  if (error != NULL)
  {
      g_warning ("%s", error->message);
      goto out;
  }
  g_cancellable_reset (cancellable);
  res = g_dbus_proxy_call_sync (self->account_proxy,
                                           "FindUserByName",
                                           g_variant_new ("(s)",
                                                          user_name),
                                           G_DBUS_CALL_FLAGS_NONE,
                                           -1,
                                           cancellable,
                                           &error);
  if (error != NULL)
  {
      g_warning ("%s", error->message);
      goto out;
  }

  g_variant_get (res, "(o)", &opath);

  self->user_path = g_strdup (opath);

  self->account_user_proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SYSTEM,
                                                G_DBUS_PROXY_FLAGS_NONE,
                                                NULL,
                                                "org.freedesktop.Accounts",
                                                self->user_path,
                                                "org.freedesktop.DBus.Properties",
                                                NULL,
                                                &error);

  if (error != NULL)
  {
      g_warning ("%s", error->message);
      goto out;
  }
  res_session= g_dbus_proxy_call_sync (self->account_user_proxy,
                                           "Get",
                                           g_variant_new ("(ss)",
                                            "org.freedesktop.Accounts.User",
                                                          "XSession"),
                                           G_DBUS_CALL_FLAGS_NONE,
                                           -1,
                                           NULL,
                                           &error);
  if (error != NULL)
  {
      g_warning ("%s", error->message);
      goto out;
  }
  g_variant_get (res_session, "(v)", &inner);
  user_session = g_variant_get_string (inner, NULL);

  if (g_strcmp0(user_session, PERFORMANCE_MODE) == 0)
  {
    self->button_state = TRUE;
  }
  else{
    self->button_state = FALSE;
  }
  gtk_switch_set_active (self->switch_btn , self->button_state);

 out:
  session_support_settings = g_settings_new (DESKTOP_SESSION_SCHEMA);
  if (g_settings_has_key (session_support_settings, SUPPORT_KEY))
    support_enabled = g_settings_get_boolean (session_support_settings, SUPPORT_KEY);
  if (support_enabled == FALSE)
  {
      gtk_widget_set_sensitive (GTK_WIDGET (self->switch_btn),  FALSE);
  }

}

