/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*-
 *
 * Copyright (C) 2012 Red Hat, Inc
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Cosimo Cecchi <cosimoc@gnome.org>
 */

#include "list-box-helper.h"
#include "cc-search-panel.h"
#include "cc-search-resources.h"
#include "cc-util.h"

#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <handy.h>

#define FULL_TEXT_SCHEMA  "org.nemo.search"
#define FULL_TEXT_KEY     "enable-file-content-search"

struct _CcSearchPanel
{
  CcPanel          parent_instance;

  GtkListBox      *search_list_box;
  GtkSwitch       *full_text_switch;
  GtkRevealer     *restart_revealer;

  GDBusProxy      *session;
  GSettings       *full_text_settings;
};

CC_PANEL_REGISTER (CcSearchPanel, cc_search_panel)

static void
restart_now (CcSearchPanel *self)
{
  g_dbus_proxy_call (self->session,
                     "Logout",
                     g_variant_new ("(u)", 0),
                     G_DBUS_CALL_FLAGS_NONE,
                     -1, NULL, NULL, NULL);
}

static void
full_text_switch_changed_cb (CcSearchPanel *self)
{
  gtk_revealer_set_reveal_child (self->restart_revealer, TRUE);
}

static void
session_proxy_ready (GObject      *source,
                     GAsyncResult *res,
                     gpointer      data)
{
  CcSearchPanel *self = data;
  GDBusProxy *proxy;
  g_autoptr(GError) error = NULL;

  proxy = g_dbus_proxy_new_for_bus_finish (res, &error);
  if (!proxy) {
    if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
      g_warning ("Failed to contact gnome-session: %s\n", error->message);
    return;
  }
  self->session = proxy;
}

static void
cc_search_panel_finalize (GObject *object)
{
  CcSearchPanel *self = CC_SEARCH_PANEL (object);

  g_clear_object (&self->full_text_settings);

  G_OBJECT_CLASS (cc_search_panel_parent_class)->finalize (object);
}

static void
cc_search_panel_class_init (CcSearchPanelClass *klass)
{
  GObjectClass *oclass = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  oclass->finalize = cc_search_panel_finalize;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/control-center/search/cc-search-panel.ui");

  gtk_widget_class_bind_template_child (widget_class, CcSearchPanel, search_list_box);
  gtk_widget_class_bind_template_child (widget_class, CcSearchPanel, full_text_switch);
  gtk_widget_class_bind_template_child (widget_class, CcSearchPanel, restart_revealer);

  gtk_widget_class_bind_template_callback (widget_class, full_text_switch_changed_cb);
  gtk_widget_class_bind_template_callback (widget_class, restart_now);
}

static void
cc_search_panel_init (CcSearchPanel *self)
{
  GSettingsSchema *schema;

  g_resources_register (cc_search_get_resource ());

  gtk_widget_init_template (GTK_WIDGET (self));

  gtk_list_box_set_header_func (self->search_list_box,
                                cc_list_box_update_header_func,
                                NULL, NULL);

  g_dbus_proxy_new_for_bus (G_BUS_TYPE_SESSION,
                            G_DBUS_PROXY_FLAGS_NONE,
                            NULL,
                            "org.gnome.SessionManager",
                            "/org/gnome/SessionManager",
                            "org.gnome.SessionManager",
                            cc_panel_get_cancellable (CC_PANEL (self)),
                            session_proxy_ready,
                            self);

  schema = g_settings_schema_source_lookup (g_settings_schema_source_get_default (), FULL_TEXT_SCHEMA, TRUE);
  if (schema)
  {
    self->full_text_settings = g_settings_new (FULL_TEXT_SCHEMA);
    if (g_settings_has_key (self->full_text_settings, FULL_TEXT_KEY))
    {
      g_settings_bind (self->full_text_settings, FULL_TEXT_KEY,
                       self->full_text_switch, "active",
                       G_SETTINGS_BIND_DEFAULT);
    }
    else
    {
      gtk_widget_set_sensitive (self->full_text_switch, FALSE);
    }
    g_settings_schema_unref (schema);
  }
  else
  {
    gtk_widget_set_sensitive (self->full_text_switch, FALSE);
  }

  gtk_revealer_set_reveal_child (self->restart_revealer, FALSE);
}
