/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2008 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <libintl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <canberra.h>
#include <handy.h>

#include <glib/gi18n-lib.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <pulse/pulseaudio.h>
#include <gvc-mixer-control.h>

#include "list-box-helper.h"
#include "cc-alert-chooser.h"
#include "cc-balance-slider.h"
#include "cc-device-combo-box.h"
#include "cc-fade-slider.h"
#include "cc-level-bar.h"
#include "cc-output-test-dialog.h"
#include "cc-profile-combo-box.h"
#include "cc-sound-panel.h"
#include "cc-sound-resources.h"
#include "cc-stream-list-box.h"
#include "cc-subwoofer-slider.h"
#include "cc-volume-slider.h"

struct _CcSoundPanel
{
  CcPanel            parent_instance;

  CcBalanceSlider   *balance_slider;
  GtkListBoxRow     *fade_row;
  CcFadeSlider      *fade_slider;
  CcDeviceComboBox  *input_device_combo_box;
  CcLevelBar        *input_level_bar;
  GtkListBox        *input_list_box;
  CcProfileComboBox *input_profile_combo_box;
  GtkListBoxRow     *input_profile_row;
  CcVolumeSlider    *input_volume_slider;
  GtkSizeGroup      *label_size_group;
  CcDeviceComboBox  *output_device_combo_box;
  GtkListStore      *output_device_model;
  CcLevelBar        *output_level_bar;
  GtkListBox        *output_list_box;
  CcProfileComboBox *output_profile_combo_box;
  GtkListBoxRow     *output_profile_row;
  CcVolumeSlider    *output_volume_slider;
  CcStreamListBox   *stream_list_box;
  GtkListBoxRow     *subwoofer_row;
  CcSubwooferSlider *subwoofer_slider;
  GtkSwitch         *noise_suppression_switch;
  HdyActionRow      *noise_suppression_row;
  gboolean          *noise_suppression_state;
  GtkListBox        *sound_effects_box;
  ca_context        *sound_context;
  GtkSwitch         *volume_enhance_switch;

  GvcMixerControl   *mixer_control;
  GSettings         *sound_settings;
  GSettings         *cdos_sound_settings;
  GSettings         *desktop_sound_settings;
  guint              timer_id;
};

CC_PANEL_REGISTER (CcSoundPanel, cc_sound_panel)

enum
{
  PROP_0,
  PROP_PARAMETERS
};

#define KEY_SOUNDS_SCHEMA "org.gnome.desktop.sound"

static void setup_noise_suppression_switch (CcSoundPanel *self);
static void setup_noise_suppression_state  (CcSoundPanel *self);

static void
allow_amplified_changed_cb (CcSoundPanel *self)
{
  cc_volume_slider_set_is_amplified (self->output_volume_slider,
                                     g_settings_get_boolean (self->sound_settings, "allow-volume-above-100-percent"));
}

static gboolean
handle_null_stream (CcSoundPanel *self)
{
  GvcMixerUIDevice *device =cc_device_combo_box_get_device (self->output_device_combo_box);
  GvcMixerStream *stream = gvc_mixer_control_get_default_sink (self->mixer_control);
  GvcChannelMap *map = gvc_mixer_stream_get_channel_map (stream);;

  cc_volume_slider_set_stream (self->output_volume_slider, stream, CC_STREAM_TYPE_OUTPUT);
  cc_level_bar_set_stream (self->output_level_bar, stream, CC_STREAM_TYPE_OUTPUT);
  cc_fade_slider_set_channel_map (self->fade_slider, map);
  cc_balance_slider_set_channel_map (self->balance_slider, map);
  cc_subwoofer_slider_set_channel_map (self->subwoofer_slider, map);
  gtk_widget_set_visible (GTK_WIDGET (self->fade_row), gvc_channel_map_can_fade (map));
  gtk_widget_set_visible (GTK_WIDGET (self->subwoofer_row), gvc_channel_map_has_lfe (map));
  gvc_mixer_control_change_output (self->mixer_control, device);

  return FALSE;
}

static void
output_device_changed_cb (CcSoundPanel *self)
{
  GvcMixerUIDevice *device;
  GvcMixerStream *stream = NULL;
  GvcChannelMap *map = NULL;
  gboolean can_fade = FALSE, has_lfe = FALSE;

  device = cc_device_combo_box_get_device (self->output_device_combo_box);

  if (device != NULL)
  {
    stream = gvc_mixer_control_get_stream_from_device (self->mixer_control, device);
    if (stream == NULL)
      g_timeout_add (100, (GSourceFunc) handle_null_stream, self);
  }

  cc_volume_slider_set_stream (self->output_volume_slider, stream, CC_STREAM_TYPE_OUTPUT);
  cc_level_bar_set_stream (self->output_level_bar, stream, CC_STREAM_TYPE_OUTPUT);

  if (stream != NULL)
    {
      map = (GvcChannelMap *) gvc_mixer_stream_get_channel_map (stream);
      can_fade = gvc_channel_map_can_fade (map);
      has_lfe = gvc_channel_map_has_lfe (map);
    }
  cc_fade_slider_set_channel_map (self->fade_slider, map);
  cc_balance_slider_set_channel_map (self->balance_slider, map);
  cc_subwoofer_slider_set_channel_map (self->subwoofer_slider, map);

  gtk_widget_set_visible (GTK_WIDGET (self->fade_row), can_fade);
  gtk_widget_set_visible (GTK_WIDGET (self->subwoofer_row), has_lfe);

  if (device != NULL)
    gvc_mixer_control_change_output (self->mixer_control, device);
}

static void
input_device_changed_cb (CcSoundPanel *self)
{
  GvcMixerUIDevice *device;
  GvcMixerStream *stream = NULL;

  device = cc_device_combo_box_get_device (self->input_device_combo_box);

  if (device != NULL)
    stream = gvc_mixer_control_get_stream_from_device (self->mixer_control, device);

  cc_volume_slider_set_stream (self->input_volume_slider, stream, CC_STREAM_TYPE_INPUT);
  cc_level_bar_set_stream (self->input_level_bar, stream, CC_STREAM_TYPE_INPUT);

  if (device != NULL)
    gvc_mixer_control_change_input (self->mixer_control, device);

  setup_noise_suppression_switch (self);
}

static void
output_device_update_cb (CcSoundPanel *self,
                         guint         id)
{
  GvcMixerUIDevice *device;
  gboolean has_multi_profiles;

  device = cc_device_combo_box_get_device (self->output_device_combo_box);
  cc_profile_combo_box_set_device (self->output_profile_combo_box, self->mixer_control, device);
  has_multi_profiles = (cc_profile_combo_box_get_profile_count (self->output_profile_combo_box) > 1);
  gtk_widget_set_visible (GTK_WIDGET (self->output_profile_row),
                          has_multi_profiles);
}

static void
input_device_update_cb (CcSoundPanel *self,
                         guint         id)
{
  GvcMixerUIDevice *device;
  gboolean has_multi_profiles;

  device = cc_device_combo_box_get_device (self->input_device_combo_box);
  cc_profile_combo_box_set_device (self->input_profile_combo_box, self->mixer_control, device);
  has_multi_profiles = (cc_profile_combo_box_get_profile_count (self->input_profile_combo_box) > 1);
  gtk_widget_set_visible (GTK_WIDGET (self->input_profile_row),
                          has_multi_profiles);
}

static void
test_output_configuration_button_clicked_cb (CcSoundPanel *self)
{
  GvcMixerUIDevice *device;
  GvcMixerStream *stream = NULL;
  CcOutputTestDialog *dialog;

  device = cc_device_combo_box_get_device (self->output_device_combo_box);
  if (device != NULL)
    stream = gvc_mixer_control_get_stream_from_device (self->mixer_control, device);

  dialog = cc_output_test_dialog_new (device, stream);
  gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (self))));
  gtk_dialog_run (GTK_DIALOG (dialog));
  gtk_widget_destroy (GTK_WIDGET (dialog));
}

static const char *
cc_sound_panel_get_help_uri (CcPanel *panel)
{
  return "help:gnome-help/media#sound";
}

static void
change_sound_file (GtkFileChooserButton *widget, gpointer user_data)
{
  GSettings * settings = user_data;
  char * filename = gtk_file_chooser_get_filename(widget);
  g_settings_set_string(settings, g_object_get_data (G_OBJECT (widget), "gsetting_key"), filename);
  g_free(filename);
}


static void
toggle_sound_file (GtkToggleButton *widget, gpointer user_data)
{
  GSettings *settings = user_data;
  gboolean active = gtk_toggle_button_get_active (widget);
  g_settings_set_boolean(settings, g_object_get_data (G_OBJECT (widget), "gsetting_key"), active);
  gtk_widget_set_sensitive(g_object_get_data (G_OBJECT (widget), "widget"), active);
}

static void
finish_callback (ca_context *c, uint32_t id, int error_code, void *user_data)
{
  g_object_set_data (G_OBJECT (user_data), "playing", "ON");
}

static void
test_sound_file (GtkButton *widget, CcSoundPanel *self)
{
  char *playing;
  char * filename = gtk_file_chooser_get_filename(g_object_get_data (G_OBJECT (widget), "widget"));    
  playing = g_object_get_data (G_OBJECT (widget), "playing");
  if (!g_strcmp0 (playing, "ON"))
  {
    ca_proplist *proplist = NULL;
    ca_proplist_create (&proplist);
    ca_proplist_sets (proplist, CA_PROP_MEDIA_FILENAME, filename);
    g_object_set_data (G_OBJECT (widget), "playing", "OFF");
    ca_context_play_full (self->sound_context, 0, proplist, finish_callback, widget);
  }  
}

static void
add_sound_effect_selector (char * label, GSettings * settings, char * file_key, char * enabled_key, CcSoundPanel *self)
{
  GtkListBoxRow *row = gtk_list_box_row_new ();
  GtkBox *box = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 12);
  gtk_widget_set_margin_bottom (box, 12);
  gtk_widget_set_margin_top(box, 12);
  gtk_widget_set_margin_right (box, 12);
  gtk_widget_set_margin_left (box, 12);
  gtk_container_add (row, box);

  GtkWidget * file_chooser = gtk_file_chooser_button_new(_("Choose an audio file"), GTK_FILE_CHOOSER_ACTION_OPEN);
  g_object_set_data (G_OBJECT (file_chooser), "gsetting_key", file_key);
  gtk_file_chooser_set_filename(file_chooser, g_settings_get_string(settings, file_key));
  gtk_widget_set_sensitive(file_chooser, g_settings_get_boolean(settings, enabled_key));
  g_signal_connect (file_chooser, "file-set", G_CALLBACK (change_sound_file), settings);
  
  GtkWidget * checkbox = gtk_check_button_new_with_label(label);
  g_object_set_data (G_OBJECT (checkbox), "widget", file_chooser);
  g_object_set_data (G_OBJECT (checkbox), "gsetting_key", enabled_key);
  gtk_widget_set_size_request (checkbox, 120, -1);
  gtk_toggle_button_set_active(checkbox, g_settings_get_boolean(settings, enabled_key));
  g_signal_connect (checkbox, "toggled", G_CALLBACK (toggle_sound_file), settings);

  GtkWidget * button = gtk_button_new();        
  GtkWidget * image = gtk_image_new_from_stock(GTK_STOCK_MEDIA_PLAY, GTK_ICON_SIZE_MENU);
  gtk_button_set_image(button, image);
  g_object_set_data (G_OBJECT (button), "widget", file_chooser);
  g_object_set_data (G_OBJECT (button), "playing", "ON");
  g_signal_connect (button, "released", G_CALLBACK (test_sound_file), self);

  gtk_box_pack_end (box, button, FALSE, FALSE, 0);
  gtk_box_pack_end (box, file_chooser, TRUE, TRUE, 0);
  gtk_box_pack_end (box, checkbox, FALSE, FALSE, 0);
  gtk_list_box_insert (self->sound_effects_box, row, -1);
}

static gboolean
setup_sound_effects (CcSoundPanel *self)
{
  ca_context_create (&self->sound_context);
  ca_context_change_props (self->sound_context, CA_PROP_APPLICATION_NAME, "cdos", CA_PROP_APPLICATION_ID, "org.Cinnamon", NULL);
  ca_context_open (self->sound_context);

  self->cdos_sound_settings = g_settings_new("org.cdos.sounds");
  self->desktop_sound_settings = g_settings_new("org.cdos.desktop.sound");

  add_sound_effect_selector (_("Starting Cdos:"), self->cdos_sound_settings, "login-file", "login-enabled", self);
  add_sound_effect_selector (_("Switching workspace:"), self->cdos_sound_settings, "switch-file", "switch-enabled", self); 
// add_sound_effect_selector (_("Mapping windows:"), self->cdos_sound_settings, "map-file", "map-enabled", self);
  add_sound_effect_selector (_("Closing windows:"), self->cdos_sound_settings, "close-file", "close-enabled", self);
  add_sound_effect_selector (_("Minimizing windows:"), self->cdos_sound_settings, "minimize-file", "minimize-enabled", self); 
  add_sound_effect_selector (_("Maximizing windows:"), self->cdos_sound_settings, "maximize-file", "maximize-enabled", self);
  add_sound_effect_selector (_("Unmaximizing windows:"), self->cdos_sound_settings, "unmaximize-file", "unmaximize-enabled", self);
//*功能未生效，移除
// add_sound_effect_selector (_("Tiling and snapping windows:"), self->cdos_sound_settings, "tile-file", "tile-enabled", self);
  add_sound_effect_selector (_("Inserting a device:"), self->cdos_sound_settings, "plug-file", "plug-enabled", self); 
  add_sound_effect_selector (_("Removing a device:"), self->cdos_sound_settings, "unplug-file", "unplug-enabled", self);
  add_sound_effect_selector (_("Changing the sound volume:"), self->desktop_sound_settings, "volume-sound-file", "volume-sound-enabled", self);
  add_sound_effect_selector (_("Leaving Cdos:"), self->cdos_sound_settings, "logout-file", "logout-enabled", self);

  gtk_widget_show_all (self->sound_effects_box);

  self->timer_id = 0;

  return FALSE;
}

static void
setup_noise_suppression_state (CcSoundPanel *self)
{
  if (0 != access ("/usr/bin/pactl", F_OK))
  {
    gtk_widget_hide (self->noise_suppression_row);
    return;
  }

  FILE *stream;
  stream = popen ("pactl list sources | grep echoCancelSource", "r");
  char strline[512];
  fgets (strline, sizeof (strline), stream);
  pclose (stream);

  if (strstr (strline, "echoCancelSource"))
    self->noise_suppression_state = TRUE;
  else
    self->noise_suppression_state = FALSE;
}

static void
noise_suppression_changed_cb (CcSoundPanel *self)
{
  GError *error = NULL;
  gchar *command = NULL;
  
  self->noise_suppression_state = gtk_switch_get_active (self->noise_suppression_switch);

  if (self->noise_suppression_state)
    command = "/usr/share/found-control-center/echoCancelTrigger.sh";
  else
    command = "/usr/share/found-control-center/echoCancelTrigger.sh off";

  g_spawn_command_line_async (command, &error);

  if (error)
    g_warning ("Failed to spawn %s: %s", command, error->message);
}

static void
setup_noise_suppression_switch (CcSoundPanel *self)
{
  if (0 != access ("/usr/share/found-control-center/echoCancelTrigger.sh", F_OK))
  {
    gtk_widget_hide (self->noise_suppression_row);
    return;
  }

  setup_noise_suppression_state (self);

  g_signal_handlers_block_by_func (self->noise_suppression_switch, noise_suppression_changed_cb, self);
  gtk_switch_set_active (self->noise_suppression_switch, self->noise_suppression_state);
  g_signal_handlers_unblock_by_func (self->noise_suppression_switch, noise_suppression_changed_cb, self);
}

static void
cc_sound_panel_finalize (GObject *object)
{
  CcSoundPanel *panel = CC_SOUND_PANEL (object);

  g_clear_object (&panel->mixer_control);
  g_clear_object (&panel->sound_settings);
  g_clear_object (&panel->cdos_sound_settings);
  g_clear_object (&panel->desktop_sound_settings);

  if (panel->timer_id != 0)
  {
    g_source_remove (panel->timer_id);
    panel->timer_id = 0;
  }

  G_OBJECT_CLASS (cc_sound_panel_parent_class)->finalize (object);
}

static void
cc_sound_panel_class_init (CcSoundPanelClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
  CcPanelClass *panel_class = CC_PANEL_CLASS (klass);

  panel_class->get_help_uri = cc_sound_panel_get_help_uri;

  object_class->finalize = cc_sound_panel_finalize;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/control-center/sound/cc-sound-panel.ui");

  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, balance_slider);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, fade_row);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, fade_slider);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, input_device_combo_box);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, input_level_bar);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, input_list_box);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, input_profile_combo_box);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, input_profile_row);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, input_volume_slider);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, label_size_group);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, output_device_combo_box);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, output_level_bar);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, output_list_box);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, output_profile_combo_box);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, output_profile_row);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, output_volume_slider);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, stream_list_box);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, subwoofer_row);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, subwoofer_slider);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, noise_suppression_switch);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, noise_suppression_row);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, sound_effects_box);
  gtk_widget_class_bind_template_child (widget_class, CcSoundPanel, volume_enhance_switch);

  gtk_widget_class_bind_template_callback (widget_class, input_device_changed_cb);
  gtk_widget_class_bind_template_callback (widget_class, output_device_changed_cb);
  gtk_widget_class_bind_template_callback (widget_class, test_output_configuration_button_clicked_cb);
  gtk_widget_class_bind_template_callback (widget_class, noise_suppression_changed_cb);

  g_type_ensure (CC_TYPE_ALERT_CHOOSER);
  g_type_ensure (CC_TYPE_BALANCE_SLIDER);
  g_type_ensure (CC_TYPE_DEVICE_COMBO_BOX);
  g_type_ensure (CC_TYPE_FADE_SLIDER);
  g_type_ensure (CC_TYPE_LEVEL_BAR);
  g_type_ensure (CC_TYPE_PROFILE_COMBO_BOX);
  g_type_ensure (CC_TYPE_STREAM_LIST_BOX);
  g_type_ensure (CC_TYPE_SUBWOOFER_SLIDER);
  g_type_ensure (CC_TYPE_VOLUME_SLIDER);
}

static void
cc_sound_panel_init (CcSoundPanel *self)
{
  g_resources_register (cc_sound_get_resource ());

  gtk_widget_init_template (GTK_WIDGET (self));

  gtk_list_box_set_header_func (self->input_list_box,
                                cc_list_box_update_header_func,
                                NULL, NULL);
  gtk_list_box_set_header_func (self->output_list_box,
                                cc_list_box_update_header_func,
                                NULL, NULL);
  gtk_list_box_set_header_func (GTK_LIST_BOX (self->stream_list_box),
                                cc_list_box_update_header_func,
                                NULL, NULL);

  self->sound_settings = g_settings_new (KEY_SOUNDS_SCHEMA);
  g_signal_connect_object (self->sound_settings,
                           "changed::allow-volume-above-100-percent",
                           G_CALLBACK (allow_amplified_changed_cb),
                           self,
                           G_CONNECT_SWAPPED);
  allow_amplified_changed_cb (self);

  self->mixer_control = gvc_mixer_control_new ("GNOME Settings");
  gvc_mixer_control_open (self->mixer_control);

  cc_stream_list_box_set_mixer_control (self->stream_list_box, self->mixer_control);
  cc_volume_slider_set_mixer_control (self->input_volume_slider, self->mixer_control);
  cc_volume_slider_set_mixer_control (self->output_volume_slider, self->mixer_control);
  cc_subwoofer_slider_set_mixer_control (self->subwoofer_slider, self->mixer_control);
  cc_device_combo_box_set_mixer_control (self->input_device_combo_box, self->mixer_control, FALSE);
  cc_device_combo_box_set_mixer_control (self->output_device_combo_box, self->mixer_control, TRUE);
  g_signal_connect_object (self->mixer_control,
                           "active-output-update",
                           G_CALLBACK (output_device_update_cb),
                           self,
                           G_CONNECT_SWAPPED);
  g_signal_connect_object (self->mixer_control,
                           "active-input-update",
                           G_CALLBACK (input_device_update_cb),
                           self,
                           G_CONNECT_SWAPPED);

  g_settings_bind (self->sound_settings, "allow-volume-above-100-percent",
                   self->volume_enhance_switch, "active",
                   G_SETTINGS_BIND_DEFAULT);

  setup_noise_suppression_switch (self);
  self->timer_id = g_timeout_add (500, (GSourceFunc) setup_sound_effects, self);
}
