/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 *
 * Copyright 2009-2010  Red Hat, Inc,
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Written by: Matthias Clasen <mclasen@redhat.com>
 */

#include "config.h"
#include <glib.h>

#include "cc-theme-info.h"

struct _CcThemeInfo {
    GObject     parent_instance;
    CcThemeType theme_type;
    GHashTable  *info_details;
};

typedef struct
{
    gchar *name;
    gchar *index_theme_file;
    GKeyFile *key_file;

} CcThemeDetail;

G_DEFINE_TYPE (CcThemeInfo, cc_theme_info, G_TYPE_OBJECT);

#define THEMES_DIR "/usr/share/themes"
#define ICONS_DIR "/usr/share/icons"

static void
cc_theme_detail_free (CcThemeDetail *detail)
{
    g_free (detail->name);
    g_free (detail->index_theme_file);
    g_key_file_unref (detail->key_file);
    g_free (detail);
}

static const gchar *
cc_theme_info_get_theme_dir (CcThemeInfo *self)
{
    const gchar *dir = NULL;
    switch (self->theme_type)
    {
        case CC_THEME_DESKTOP:
        case CC_THEME_COLOR:
            dir = THEMES_DIR;
            break;
        case CC_THEME_ICONS:
            dir = ICONS_DIR;
            break;
        case CC_THEME_CURSOR:
            dir = ICONS_DIR;
            break;
        default:
            break;
    }
    return dir;
}

static const gchar *
cc_theme_info_get_theme_group (CcThemeInfo *self)
{
    const gchar *group = NULL;
    switch (self->theme_type)
    {
        case CC_THEME_DESKTOP:
        case CC_THEME_COLOR:
            group = DESKTOP_ENTRY_GROUP;
            break;
        case CC_THEME_ICONS:
        case CC_THEME_CURSOR:
            group = ICON_THEME_GROUP;
            break;
        default:
            break;
    }
    return group;
}

static const gchar *
cc_theme_info_get_index_theme_name (CcThemeInfo *self)
{
    const gchar *index_theme = NULL;
    switch (self->theme_type)
    {
        case CC_THEME_DESKTOP:
        case CC_THEME_ICONS:
        case CC_THEME_COLOR:
            index_theme = "index.theme";
            break;
        case CC_THEME_CURSOR:
            index_theme = "cursor.theme";
            break;
        default:
            break;
    }
    return index_theme;
}

static gboolean
check_desktop_theme_file (CcThemeInfo *self, GKeyFile *key_file)
{
    gchar *type = NULL;
    gboolean ret = TRUE;
    const gchar *group = cc_theme_info_get_theme_group (self);
    if (g_key_file_has_group (key_file, group) == FALSE)
        return FALSE;

    type = g_key_file_get_string (key_file, 
                                  group, 
                                  "Type", 
                                  NULL);
    if (type == NULL || g_strcmp0 (type, DESKTOP_THEME_TYPE))
        ret = FALSE;
    g_free (type);

    return ret;
}

static gboolean
check_color_theme_file (CcThemeInfo *self, GKeyFile *key_file)
{
    gchar *type = NULL;
    gboolean ret = TRUE;
    const gchar *group = cc_theme_info_get_theme_group (self);
    if (g_key_file_has_group (key_file, group) == FALSE)
        return FALSE;

    type = g_key_file_get_string (key_file, 
                                  group, 
                                  "Type", 
                                  NULL);
    //桌面默认主题也是颜色主题的一种
    if (type == NULL || (g_strcmp0 (type, COLOR_THEME_TYPE) && g_strcmp0 (type, DESKTOP_THEME_TYPE)))
        ret = FALSE;
    g_free (type);

    return ret;
}

static gboolean
check_icons_theme_file (CcThemeInfo *self, GKeyFile *key_file)
{
    gchar *type = NULL;
    gboolean ret = TRUE;
    const gchar *group = cc_theme_info_get_theme_group (self);
    if (g_key_file_has_group (key_file, group) == FALSE)
        return FALSE;
    type = g_key_file_get_string (key_file, 
                                  group, 
                                  "Type", 
                                  NULL);
    if (type == NULL || g_strcmp0 (type, ICON_THEME_TYPE))
        ret = FALSE;
    g_free (type);

    return ret;
}

static gboolean
check_cursor_theme_file (CcThemeInfo *self, GKeyFile *key_file)
{
    gchar *type = NULL;
    gboolean ret = TRUE;
    const gchar *group = cc_theme_info_get_theme_group (self);
    if (g_key_file_has_group (key_file, group) == FALSE)
        return FALSE;
    type = g_key_file_get_string (key_file, 
                                  group, 
                                  "Type", 
                                  NULL);
    if (type == NULL || g_strcmp0 (type, CURSOR_THEME_TYPE))
        ret = FALSE;
    g_free (type);

    return ret;
}

static gboolean
cc_theme_info_check_index_theme_file (CcThemeInfo *self, GKeyFile *key_file)
{
    gboolean ret = FALSE;
    switch (self->theme_type)
    {
        case CC_THEME_DESKTOP:
            ret = check_desktop_theme_file (self, key_file);
            break;
        case CC_THEME_COLOR:
            ret = check_color_theme_file (self, key_file);
            break;
        case CC_THEME_ICONS:
            ret = check_icons_theme_file (self, key_file);
            break;
        case CC_THEME_CURSOR:
            ret = check_cursor_theme_file (self, key_file);
            break;
        default:
            break;
    }
    return ret;
}

static void
cc_theme_info_add_detail (CcThemeInfo *self, gchar *theme_path)
{
    gchar *index_file;
    CcThemeDetail *detail;
    GKeyFile *key_file = NULL;
    GError *error = NULL;
    gboolean is_valid = FALSE;
    const gchar *index_theme_name;

    index_theme_name = cc_theme_info_get_index_theme_name (self);
    if (!index_theme_name)
        return;
    index_file = g_build_filename (theme_path, index_theme_name, NULL);
    if (!g_file_test (index_file, G_FILE_TEST_EXISTS))
        goto out;

    key_file = g_key_file_new ();
    g_key_file_load_from_file (key_file, 
                               index_file, 
                               G_KEY_FILE_NONE, 
                               &error);
    if (error)
    {
        g_warning ("Load the index.theme file failed: %s", error->message);
        g_error_free (error);
        goto out;
    }

    //判断是否是有效的主题文件
    is_valid = cc_theme_info_check_index_theme_file (self, key_file);
    if (!is_valid)
        goto out;

    detail = g_new0 (CcThemeDetail, 1);
    detail->name = g_filename_display_basename (theme_path);
    detail->index_theme_file = g_strdup (index_file);
    detail->key_file = g_key_file_ref (key_file);
    g_hash_table_insert (self->info_details, 
                         g_strdup (detail->name), 
                         detail);
    g_debug ("%s:: insert key = %s, theme_path = %s", __func__, detail->name, theme_path);
out:
    g_free (index_file);
    if (key_file)
        g_key_file_unref (key_file);
}

static void
cc_theme_info_fetch_themes (CcThemeInfo *self)
{
    const gchar *dir_path;
    const gchar *theme_name;
    gchar *theme_path;
    GError *error = NULL;
    GDir *dir;

    dir_path = cc_theme_info_get_theme_dir (self);
    if (!dir_path)
        return;

    g_debug ("%s:: dir_path = %s", __func__, dir_path);
    dir = g_dir_open (dir_path, 0, &error);
    if (error)
    {
        g_warning ("Open theme dir failed: %s", error->message);
        g_error_free (error);
        return;
    }

    while (TRUE)
    {
        theme_name = g_dir_read_name (dir);
        if (theme_name == NULL)
            break;
        theme_path = g_build_filename (dir_path, theme_name, NULL);
        cc_theme_info_add_detail (self, theme_path);
        g_free (theme_path);
    }
}

static void
cc_theme_info_setup (CcThemeInfo *self)
{
    cc_theme_info_fetch_themes (self);
}

static void
cc_theme_info_init (CcThemeInfo *self)
{
    self->info_details = g_hash_table_new_full (g_str_hash, 
                                                g_str_equal,
                                                g_free, 
                                                cc_theme_detail_free);
}

static void
cc_theme_info_dispose (GObject *obj)
{
    CcThemeInfo *self = CC_THEME_INFO (obj);
    if (self->info_details)
    {
        g_hash_table_unref (self->info_details);
        self->info_details = NULL;
    }
    G_OBJECT_CLASS (cc_theme_info_parent_class)->dispose (obj);
}

static void
cc_theme_info_finalize (GObject *obj)
{
    CcThemeInfo *self = CC_THEME_INFO (obj);
    G_OBJECT_CLASS (cc_theme_info_parent_class)->finalize (obj);
}

static void
cc_theme_info_class_init (CcThemeInfoClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->dispose = cc_theme_info_dispose;
    object_class->finalize = cc_theme_info_finalize;
}

CcThemeInfo *
cc_theme_info_new (CcThemeType theme_type)
{
    CcThemeInfo *self;
    self = (CcThemeInfo *)g_object_new (CC_TYPE_THEME_INFO, NULL);
    self->theme_type = theme_type;
    cc_theme_info_setup (self);
    return self;
}

GList *
cc_theme_info_get_all (CcThemeInfo *self)
{
    g_return_val_if_fail (CC_IS_THEME_INFO (self), NULL);
    return g_hash_table_get_keys (self->info_details);
}

GKeyFile *
cc_theme_info_get_key_file (CcThemeInfo *self, gchar *name)
{
    g_return_val_if_fail (CC_IS_THEME_INFO (self), NULL);
    g_return_val_if_fail (name != NULL, NULL);
    GKeyFile *key_file;
    CcThemeDetail *detail;
    detail = g_hash_table_lookup (self->info_details, name);
    if (detail)
        return detail->key_file;
    return NULL;
}

gchar *
cc_theme_info_get_display_name (CcThemeInfo *self, gchar *name)
{
    g_return_val_if_fail (CC_IS_THEME_INFO (self), NULL);
    g_return_val_if_fail (name != NULL, NULL);
    GKeyFile *key_file;
    const gchar *group;
    gchar *display_name;

    group = cc_theme_info_get_theme_group (self);
    key_file = cc_theme_info_get_key_file (self, name);
    if (!group || !key_file)
        return NULL;
    display_name = g_key_file_get_locale_string (key_file, 
                                                 group, 
                                                 "Name", 
                                                 NULL, 
                                                 NULL);
    return display_name;
}

gchar *
cc_theme_info_get_comment (CcThemeInfo *self, gchar *name)
{
    g_return_val_if_fail (CC_IS_THEME_INFO (self), NULL);
    g_return_val_if_fail (name != NULL, NULL);
    GKeyFile *key_file;
    const gchar *group;
    gchar *comment;

    group = cc_theme_info_get_theme_group (self);
    key_file = cc_theme_info_get_key_file (self, name);
    if (!group || !key_file)
        return NULL;
    comment = g_key_file_get_locale_string (key_file, 
                                            group, 
                                            "Comment", 
                                            NULL, 
                                            NULL);
    return comment;
}

/* 判断是否为有效的主题 */
gboolean 
cc_theme_info_is_valid_theme (CcThemeInfo *self, gchar *name)
{
    g_return_val_if_fail (CC_IS_THEME_INFO (self), FALSE);
    g_return_val_if_fail (name != NULL, FALSE);
    GKeyFile *key_file;
    const gchar *group;

    group = cc_theme_info_get_theme_group (self);
    key_file = cc_theme_info_get_key_file (self, name);
    if (!group || !key_file)
        return FALSE;
    return cc_theme_info_check_index_theme_file (self, key_file);
}