/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*-
 *
 * Copyright (C) 2017 Mohammed Sadiq <sadiq@sadiqpk.org>
 * Copyright (C) 2010 Red Hat, Inc
 * Copyright (C) 2008 William Jon McCann <jmccann@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <config.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <stdio.h>
#include <dirent.h>

#include "list-box-helper.h"
#include "cc-theme-info.h"
#include "cc-theme-panel.h"
#include "cc-theme-resources.h"
#include "cc-util.h"

/* Autorun options */
#define GTK_THEME                 "gtk-theme"
#define STYLE_THEME               "style-theme"
#define COLOR_THEME               "color-theme"
#define COLOR_THEME_DARK          "color-theme-dark"
#define ICON_THEME                "icon-theme"
#define CURSOR_THEME              "cursor-theme"
#define DIN_THEME                 "change-theme-din"

#define INTERFACE_SCHEMA          "org.cdos.desktop.interface"
#define COLOR_SCHEMA              "org.cdos.settings-daemon.plugins.color"
#define MUTTER_SCHEMA             "org.gnome.mutter"

struct _CcThemePanel
{
  CcPanel              parent_instance;
  GtkWidget *          theme_list_box;
  GtkWidget *          window_list_box;
  GtkWidget *          style_theme_combox;
  GtkWidget *          color_theme_combox;
  GtkWidget *          icon_theme_combox;
  GtkWidget *          cursor_theme_combox;
  GtkSwitch *          day_night_switch;
  GtkSwitch *          effect_switch;
  GtkScale  *          transparency_scale;

  GSettings           *settings;
  GSettings           *color_settings;
  GSettings           *mutter_settings;

  guint               *settings_id;

  /* 桌面样式主题 */
  CcThemeInfo         *desktop_theme_info;
  CcThemeInfo         *color_theme_info;
  CcThemeInfo         *icon_theme_info;
  CcThemeInfo         *cursor_theme_info;
};

CC_PANEL_REGISTER (CcThemePanel, cc_theme_panel)

static void
update_combobox_with_model (CcThemePanel   *self, 
                            CcThemeInfo    *theme_info,
                            GtkWidget      *combobox,
                            const gchar    *key);

static void
ellipsize_cell_layout (GtkCellLayout *cell_layout)
{
  GList *cells = gtk_cell_layout_get_cells (cell_layout);
  GList *cell;

  for (cell = cells; cell; cell = cell->next)
    if (GTK_IS_CELL_RENDERER_TEXT (cell->data))
      g_object_set (G_OBJECT (cell->data), "ellipsize", PANGO_ELLIPSIZE_END, NULL);
  g_list_free (cells);
}

/* 获取index.theme文件内定义MetaTheme
 * MetaTheme对应原生样式主题，此种主题可以衍生出多个配色的独立主题
 * 当前桌面的MetaTheme有Found和FoundDark两种，分别对应常规色系和暗色系
 * 桌面样式主题也是以此作为判断标准
 */
static gchar *
theme_info_get_meta_theme_name (CcThemePanel *self, gchar *name)
{
  gchar *style_theme = NULL;
  GKeyFile *key_file;

  key_file = cc_theme_info_get_key_file (self->color_theme_info, name);
  if (key_file)
  {
    style_theme = g_key_file_get_string (key_file, 
                                         DESKTOP_THEME_GROUP, 
                                         "MetaTheme", 
                                         NULL);
  }
  return style_theme;
}

/* 
 * 获取主题色调定义
 * ColorHue常用于判断是否属于暗色系样式主题，如果没有此值，则认为是非按色系的
 * 
 */
static gchar *
theme_info_get_color_hue (CcThemePanel *self, gchar *name)
{
  gchar *color_hue = NULL;
  GKeyFile *key_file;

  key_file = cc_theme_info_get_key_file (self->color_theme_info, name);
  if (key_file)
  {
    color_hue = g_key_file_get_string (key_file, 
                                       DESKTOP_THEME_GROUP, 
                                       "ColorHue", 
                                       NULL);
  }
  return color_hue;
}

/* 
 * 获取图标主题
 * 例如暗色主题配暗色图标
 */
static gchar *
theme_info_get_icon_theme (CcThemePanel *self, gchar *name)
{
  gchar *icon_theme = NULL;
  GKeyFile *key_file;

  key_file = cc_theme_info_get_key_file (self->color_theme_info, name);
  if (key_file)
  {
    icon_theme = g_key_file_get_string (key_file, 
                                       DESKTOP_THEME_GROUP, 
                                       "IconTheme", 
                                       NULL);
  }
  return icon_theme;
}

/* 是否为暗色主题 */
static gboolean
theme_info_is_dark_style (CcThemePanel *self, gchar *name)
{
  gchar *color_hue;
  gboolean is_dark = FALSE;
  color_hue = theme_info_get_color_hue (self, name);
  if (color_hue && g_strcmp0 (color_hue, "dark") == 0)
    is_dark = TRUE;
  g_free (color_hue);
  return is_dark;
}

/* 根据GTK_THEME获取对应的样式主题 */
static gchar *
get_style_theme_from_gsettings (CcThemePanel *self)
{
  gchar *gtk_theme, *style_theme = NULL;
  GKeyFile *key_file;

  gtk_theme = g_settings_get_string (self->settings, GTK_THEME);
  style_theme = theme_info_get_meta_theme_name (self, gtk_theme);
  g_free (gtk_theme);
  return style_theme;
}

static void
style_theme_changed (GtkComboBox *widget, CcThemePanel *self)
{
  GtkTreeIter   iter;
  gchar        *theme = NULL, *gtk_theme, *color_theme = NULL;
  GtkTreeModel *model;
  gboolean      is_dark;
  gchar        *icon_theme;

  if (gtk_combo_box_get_active_iter (widget, &iter))
  {
    model = gtk_combo_box_get_model (widget);
    gtk_tree_model_get (model, &iter, 1, &theme, -1);
    gtk_theme = g_settings_get_string (self->settings, GTK_THEME);

    /* 缓存当前主题，以便下次还原
     * 暗色主题缓存到COLOR_THEME_DARK
     * 常规主题缓存到COLOR_THEME
     */
    is_dark = theme_info_is_dark_style (self, gtk_theme);
    g_debug ("%s:%d: is_dark = %d, gtk_theme = %s", __func__, __LINE__, is_dark, gtk_theme);
    if (is_dark)
    {
      g_settings_set_string (self->settings, COLOR_THEME_DARK, gtk_theme);
    } else {
      g_settings_set_string (self->settings, COLOR_THEME, gtk_theme);
    }

    /*更新当前样式主题对应的颜色主题
      注意：从缓存的gsettings内获取(COLOR_THEME和COLOR_THEME_DARK)。
           如果获取不到就用当前的。
    */
    is_dark = theme_info_is_dark_style (self, theme);
    g_debug ("%s:%d: is_dark = %d, theme = %s", __func__, __LINE__, is_dark, theme);
    if (is_dark)
    {
      color_theme = g_settings_get_string (self->settings, COLOR_THEME_DARK);
    } else {
      color_theme = g_settings_get_string (self->settings, COLOR_THEME);
    }

    if (g_strcmp0 (color_theme, "") == 0)
    {
      g_clear_pointer (&color_theme, g_free);
    }

    /* 获取选中主题文件的index.theme中定义MetaTheme */
    if (color_theme == NULL)
      color_theme = theme_info_get_meta_theme_name (self, theme); 

    if (color_theme == NULL)
        color_theme = g_strdup (theme);

    g_debug ("%s::set new color theme: %s", __func__, color_theme);
    /* 如果不是有效的主题，则重置为默认 */
    if (cc_theme_info_is_valid_theme (self->color_theme_info, color_theme))
    {
      g_settings_set_string (self->settings, GTK_THEME, color_theme);
      //更新图标主题，如暗色主题配暗色图标
      icon_theme = theme_info_get_icon_theme (self, theme); 
      if (icon_theme)
      {
        if (g_strcmp0 (icon_theme, "default") == 0 || g_strcmp0 (icon_theme, "Default") == 0) //重置为默认
        {
          g_settings_reset (self->settings, ICON_THEME);
        } 
        else if (g_strcmp0 (icon_theme, ""))  //跳过空白
        {
          g_settings_set_string (self->settings, ICON_THEME, icon_theme);
        }
        g_free (icon_theme);
      }
    } else {
      g_settings_reset (self->settings, GTK_THEME);
    }

    g_free (theme);
    g_free (gtk_theme);
    g_free (color_theme);
  }
}

static void
color_theme_changed (GtkComboBox *widget, CcThemePanel *self)
{
  GtkTreeIter   iter;
  gchar        *theme = NULL;
  GtkTreeModel *model;

  if (gtk_combo_box_get_active_iter (widget, &iter))
  {
    model = gtk_combo_box_get_model (widget);
    gtk_tree_model_get (model, &iter, 1, &theme, -1);
    //防止调用闭环
    //g_signal_handler_block (self->settings, self->settings_id);
    g_settings_set_string (self->settings, GTK_THEME, theme);
    //g_signal_handler_unblock (self->settings, self->settings_id);
    g_free (theme);
  }
}

static void
icon_theme_changed (GtkComboBox *widget, CcThemePanel *self)
{
  GtkTreeIter   iter;
  gchar        *theme = NULL;
  GtkTreeModel *model;

  if (gtk_combo_box_get_active_iter (widget, &iter))
  {
    model = gtk_combo_box_get_model (widget);
    gtk_tree_model_get (model, &iter, 1, &theme, -1);
    g_settings_set_string (self->settings, ICON_THEME, theme);
  }
  g_free (theme);
}

static void
cursor_theme_changed (GtkComboBox *widget, CcThemePanel *self)
{
  GtkTreeIter   iter;
  gchar        *theme = NULL;
  GtkTreeModel *model;

  if (gtk_combo_box_get_active_iter (widget, &iter))
  {
    model = gtk_combo_box_get_model (widget);
    gtk_tree_model_get (model, &iter, 1, &theme, -1);
    g_settings_set_string (self->settings, CURSOR_THEME, theme);
  }
  g_free (theme);
}

static void
update_combobox_with_model (CcThemePanel   *self, 
                            CcThemeInfo    *theme_info,
                            GtkWidget      *combobox,
                            const gchar    *key)
{
  GtkListStore *store = NULL;
  GtkTreeIter iter;
  GtkCellRenderer *renderer = NULL;
  gchar *current_theme = NULL, *style_theme, *meta_theme;
  GList *themes, *l;
  gboolean in_themes = FALSE;
  //"name" can't be free. 
  gchar *name, *display_name;

  store = (GtkListStore *)gtk_combo_box_get_model (combobox);
  if (store)
  {
    gtk_list_store_clear ((GtkListStore *)store);
    gtk_cell_layout_clear (GTK_CELL_LAYOUT(combobox));
  } else {
    store = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_STRING);
    gtk_combo_box_set_model (combobox, GTK_TREE_MODEL(store));
    gtk_combo_box_set_id_column (combobox, 1);
  }

  if (g_strcmp0 (key, STYLE_THEME) == 0)
  {
    current_theme = get_style_theme_from_gsettings (self);
    meta_theme = g_strdup (current_theme);
  } else {
    current_theme = g_settings_get_string (self->settings, key);
    meta_theme = theme_info_get_meta_theme_name (self, current_theme);
  }
  g_debug ("%s:%d: current_theme = %s, meta_theme = %s", __func__, __LINE__, current_theme, meta_theme);
  themes = cc_theme_info_get_all (theme_info);
  l = themes;
  while (l)
  {
    name = l->data;
    l = l->next;

    //只更新当前样式主题一致的颜色主题列表
    if (g_strcmp0 (key, GTK_THEME) == 0)
    {
      style_theme = theme_info_get_meta_theme_name (self, name);
      g_debug ("%s::current_theme = %s, style_theme = %s", __func__, current_theme, style_theme);
      if (g_strcmp0 (style_theme, meta_theme) != 0)
      {
        g_free (style_theme);
        continue;
      }
      g_free (style_theme);
    }

    display_name = cc_theme_info_get_display_name (theme_info, name);
    g_debug ("%s::theme = %s, display_name = %s", __func__, name, display_name);
    if (!display_name)
      display_name = g_strdup (name);

    gtk_list_store_append (store, &iter);
    gtk_list_store_set (store, &iter, 
                        0, display_name,
                        1, name, 
                        -1);
    if (current_theme && g_strcmp0 (current_theme, name) == 0)
    {
      gtk_combo_box_set_active_iter (combobox, &iter);
      in_themes = TRUE;
    }
    g_free (display_name);
  }
  g_list_free (themes);

  if (current_theme && in_themes == FALSE)
  {
    gtk_list_store_append (store, &iter);
    gtk_list_store_set (store, &iter, 0, current_theme, 1, current_theme, -1);
    gtk_combo_box_set_active_iter (combobox, &iter);
  }

  // 设置GtkTreeModel中的每一项数据如何在列表框中显示
  renderer = gtk_cell_renderer_text_new ();
  gtk_cell_layout_pack_start (GTK_CELL_LAYOUT(combobox), renderer, TRUE);
  gtk_cell_layout_set_attributes ( GTK_CELL_LAYOUT (combobox), renderer, "text", 0, NULL);

  g_free (current_theme);
  g_free (meta_theme);
}

static void
info_panel_setup_media (CcThemePanel *self)
{
  update_combobox_with_model (self, 
                              self->desktop_theme_info, 
                              self->style_theme_combox, 
                              STYLE_THEME);
  update_combobox_with_model (self, 
                              self->color_theme_info, 
                              self->color_theme_combox, 
                              GTK_THEME);
  update_combobox_with_model (self, 
                              self->icon_theme_info, 
                              self->icon_theme_combox, 
                              ICON_THEME);
  update_combobox_with_model (self, 
                              self->cursor_theme_info, 
                              self->cursor_theme_combox, 
                              CURSOR_THEME);

  g_signal_connect (self->style_theme_combox, 
                    "changed", 
                    G_CALLBACK(style_theme_changed), 
                    self);
  g_signal_connect (self->color_theme_combox, 
                    "changed", 
                    G_CALLBACK(color_theme_changed), 
                    self);
  g_signal_connect (self->icon_theme_combox, 
                    "changed", 
                    G_CALLBACK(icon_theme_changed), 
                    self);
  g_signal_connect (self->cursor_theme_combox, 
                    "changed", 
                    G_CALLBACK(cursor_theme_changed), 
                    self);
}

static void
cc_theme_panel_finalize (GObject *object)
{
  CcThemePanel *self = CC_THEME_PANEL (object);

  G_OBJECT_CLASS (cc_theme_panel_parent_class)->finalize (object);
}

static void
theme_reset_default_settings (CcThemePanel *self)
{
  g_settings_reset(self->settings, GTK_THEME);
  g_settings_reset(self->settings, ICON_THEME);
  g_settings_reset(self->settings, CURSOR_THEME);
  g_settings_reset(self->color_settings, DIN_THEME);

  if (g_settings_has_key (self->mutter_settings, "window-border"))
    g_settings_reset(self->mutter_settings, "window-border");
  if (g_settings_has_key (self->mutter_settings, "titlebar-opacity"))
    g_settings_reset(self->mutter_settings, "titlebar-opacity");
}

static void
set_active_iter (GSettings      *settings,
                 const gchar    *key, 
                 GtkWidget      *combobox)
{
  GtkTreeIter   iter;
  GtkTreeModel *model;
  gboolean      valid;
  gchar *theme_value = NULL;
  gchar *model_value = NULL;

  theme_value = g_settings_get_string (settings, key);
  if (theme_value)
  {
    model = gtk_combo_box_get_model (combobox);
    for (valid = gtk_tree_model_get_iter_first (model, &iter);
         valid;
         valid = gtk_tree_model_iter_next (model, &iter))
    {
      gtk_tree_model_get (model, &iter, 1, &model_value, -1);
      if (g_strcmp0  (theme_value, model_value) == 0)
      {
        gtk_combo_box_set_active_iter (combobox, &iter);
        break;
      }
    }  
  }
}

static void 
settings_update_active (GSettings     *settings,
                        const gchar   *key,
                        CcThemePanel  *self)
{
  //防止调用闭环
  g_signal_handler_block (settings, self->settings_id);
  if (g_strcmp0  (key, GTK_THEME) == 0)
  {
    //set_active_iter (settings, key, self->color_theme_combox);
    //刷新列表
    update_combobox_with_model (self, 
                              self->desktop_theme_info, 
                              self->style_theme_combox, 
                              STYLE_THEME);
    update_combobox_with_model (self, 
                              self->color_theme_info, 
                              self->color_theme_combox, 
                              GTK_THEME);
    update_combobox_with_model (self, 
                              self->icon_theme_info, 
                              self->icon_theme_combox, 
                              ICON_THEME);
  }
  else if (g_strcmp0  (key, ICON_THEME) == 0)
    set_active_iter (settings, key, self->icon_theme_combox);
  else if (g_strcmp0  (key, CURSOR_THEME) == 0)
    set_active_iter (settings, key, self->cursor_theme_combox);

  g_signal_handler_unblock (settings, self->settings_id);
}

static void 
graphic_effects_changed_cb(CcThemePanel *self)
{
  gboolean state = g_settings_get_boolean (self->mutter_settings, "window-border");
  g_settings_set_boolean (self->mutter_settings, "window-radius", state);
  g_settings_set_boolean (self->mutter_settings, "window-shadow", state);
}

static void
cc_theme_panel_dispose (GObject *object)
{
  CcThemePanel *self = CC_THEME_PANEL (object);
  g_clear_object (&self->settings);
  g_clear_object (&self->color_settings);
  g_clear_object (&self->mutter_settings);
  g_clear_object (&self->desktop_theme_info);
  g_clear_object (&self->color_theme_info);
  g_clear_object (&self->icon_theme_info);
  g_clear_object (&self->cursor_theme_info);
  G_OBJECT_CLASS (cc_theme_panel_parent_class)->dispose (object);
}

static void
cc_theme_panel_class_init (CcThemePanelClass *klass)
{
  GObjectClass   *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  object_class->finalize = cc_theme_panel_finalize;
  object_class->dispose = cc_theme_panel_dispose;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/control-center/theme/cc-theme-panel.ui");

  gtk_widget_class_bind_template_child (widget_class, CcThemePanel, theme_list_box);
  gtk_widget_class_bind_template_child (widget_class, CcThemePanel, window_list_box);
  gtk_widget_class_bind_template_child (widget_class, CcThemePanel, style_theme_combox);
  gtk_widget_class_bind_template_child (widget_class, CcThemePanel, color_theme_combox);
  gtk_widget_class_bind_template_child (widget_class, CcThemePanel, icon_theme_combox);
  gtk_widget_class_bind_template_child (widget_class, CcThemePanel, cursor_theme_combox);
  gtk_widget_class_bind_template_child (widget_class, CcThemePanel, day_night_switch);
  gtk_widget_class_bind_template_child (widget_class, CcThemePanel, effect_switch);
  gtk_widget_class_bind_template_child (widget_class, CcThemePanel, transparency_scale);
  gtk_widget_class_bind_template_callback (widget_class, theme_reset_default_settings);

}

static void
cc_theme_panel_init (CcThemePanel *self)
{
  g_resources_register (cc_theme_get_resource ());
  gtk_widget_init_template (GTK_WIDGET (self));
  self->settings = g_settings_new (INTERFACE_SCHEMA);
  self->color_settings = g_settings_new (COLOR_SCHEMA);
  self->mutter_settings = g_settings_new (MUTTER_SCHEMA);
  self->desktop_theme_info = cc_theme_info_new (CC_THEME_DESKTOP);
  self->color_theme_info = cc_theme_info_new (CC_THEME_COLOR);
  self->icon_theme_info = cc_theme_info_new (CC_THEME_ICONS);
  self->cursor_theme_info = cc_theme_info_new (CC_THEME_CURSOR);

  ellipsize_cell_layout (GTK_CELL_LAYOUT (self->style_theme_combox));
  ellipsize_cell_layout (GTK_CELL_LAYOUT (self->color_theme_combox));
  ellipsize_cell_layout (GTK_CELL_LAYOUT (self->icon_theme_combox));
  ellipsize_cell_layout (GTK_CELL_LAYOUT (self->cursor_theme_combox));

  gtk_list_box_set_header_func (self->theme_list_box,
                                cc_list_box_update_header_func,
                                NULL, NULL);
  gtk_list_box_set_header_func (self->window_list_box,
                                cc_list_box_update_header_func,
                                NULL, NULL);
  
  info_panel_setup_media (self);
  self->settings_id = \
  g_signal_connect (self->settings, 
                    "changed",
                    G_CALLBACK (settings_update_active), 
                    self);
  if (g_settings_has_key (self->color_settings, DIN_THEME))
  {
    g_settings_bind (self->color_settings, 
                    DIN_THEME,
                    self->day_night_switch, 
                    "active",
                    G_SETTINGS_BIND_DEFAULT);
  } else {
    gtk_widget_set_sensitive (self->day_night_switch, FALSE);
  }

  if (g_settings_has_key (self->mutter_settings, "window-border"))
  {
    g_settings_bind (self->mutter_settings, 
                    "window-border",
                    self->effect_switch, 
                    "active",
                    G_SETTINGS_BIND_DEFAULT);
    g_signal_connect_object (self->mutter_settings, 
                            "changed::window-border",
                            G_CALLBACK (graphic_effects_changed_cb), 
                            self,
                            G_CONNECT_SWAPPED);
  } else {
    gtk_widget_set_sensitive (self->effect_switch, FALSE);
  }

  if (g_settings_has_key (self->mutter_settings, "titlebar-opacity"))
  {
    g_settings_bind (self->mutter_settings, 
                    "titlebar-opacity",
                    gtk_range_get_adjustment (GTK_RANGE (self->transparency_scale)), 
                    "value",
                    G_SETTINGS_BIND_DEFAULT);
  } else {
    gtk_widget_set_sensitive (self->transparency_scale, FALSE);
  }

  if (g_settings_has_key (self->mutter_settings, "window-decoration-can-modify"))
  {
    gboolean can_modify = g_settings_get_boolean (self->mutter_settings, "window-decoration-can-modify");
    gtk_widget_set_sensitive (self->transparency_scale, can_modify);
    gtk_widget_set_sensitive (self->effect_switch, can_modify);
  }
}
